"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
from .BaseLayer import BaseLayer

class OutputLayer(BaseLayer):
    def __init__(self, activation = "relu", name = "layer"):
        BaseLayer.__init__(self, activation, name = name)

    def __deepcopy__(self, memo = None):
        pass

    def load_weights(self, **kwargs):
        self.input_shape = kwargs["input_shape"]

        self.output_shape = self.input_shape
        
        return self.output_shape
    
    def forward_pass(self, **kwargs):
        inputv = kwargs["inputv"]
        if self.activation is not None:
            return BaseLayer.activations[self.activation](inputv)
        return inputv



"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
import numpy as np, itertools
from numba import cuda
from .BaseLayer import BaseLayer
from .cuda_utils import matmul
from .cpu_utils import mul
          
class DenseLayer(BaseLayer):
    def __init__(self, activation = "relu", quant_nbits = 8, multiplier = BaseLayer.default_multiplier, name = "layer",offset=[128,128]):
        super().__init__( activation, quant_nbits, multiplier, name)
        self.offset=offset

    def __deepcopy__(self, memo = None):
        return DenseLayer(activation = self.activation, quant_nbits = self.quant_nbits, multiplier = self.multiplier, name = self.name)

    def forward_pass(self, **kwargs):
        # TODO Aggiungi supporto per altre attivazioni
        activation = self.activation == "relu"

        if self.gpu_input_memory is None:
            inputv = kwargs["inputv"]
            #print("dense no buoeno")
            A_global_mem = cuda.to_device(np.array(inputv))
        else:
            A_global_mem=self.gpu_input_memory

        cuda.synchronize()
        matmul[self.griddim, self.blockdim](self.results, A_global_mem, self.weights, self.biases, self.M, self.offset[0], self.offset[1], activation, self.quant_nbits)#offset1,offset2)
        cuda.synchronize()
        if self.gpu_output_memory == False:
            self.outputv[:,:] = self.results.copy_to_host()
        if self.activation == "softmax":
            if self.activation is not None:
                self.outputv = BaseLayer.activations[self.activation](self.outputv)
            if self.quant_nbits is not None:
                self.outputv = np.round((self.outputv/np.max(self.outputv))*(2**self.quant_nbits-1))
            return self.outputv

    def load_weights(self, **kwargs):
        self.enable_gpu = kwargs["enable_gpu"]
        self.input_shape = kwargs["input_shape"]
        self.weights, self.biases =kwargs["weights"], kwargs["biases"]
        self.output_shape = [self.input_shape[0], kwargs["weights"].shape[1]]     
        self.outputv = np.zeros(self.output_shape)
        self.griddim = (self.output_shape[0] // self.blockdim[0] + 1, self.output_shape[1] // self.blockdim[1] + 1)#,n_channels)
        if self.enable_gpu:
            self.use_gpu=True
            self.M = self.multiplier
            self.weights = cuda.to_device(np.array(self.weights))
            self.biases = cuda.to_device(np.array(self.biases))
            self.results = cuda.device_array(self.output_shape)
            self.gpu_input_memory = kwargs["gpu_input_memory"]
        if self.gpu_input_memory is None:
            self.outputv = np.zeros(self.output_shape)   
        return self.output_shape
   
    def dense_kernel(self,neurons, output_shape, inputv, weights, multiplier):
        for j, i in itertools.product(neurons, range(output_shape[0])):
            self.outputv[i, j] = np.sum(mul(inputv[i,:], weights[:,j], multiplier))
        return self.outputv
    



"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""

import numpy as np, json, copy, tensorflow as tf, time
from distutils.dir_util import mkpath
from .ConvLayer import ConvLayer
from .DenseLayer import DenseLayer
from .PoolingLayer import PoolingLayer
from .FlattenLayer import FlattenLayer
from .OutputLayer import OutputLayer
from .utils import grouped, flatten


class NetworkModel:
    def __init__(self, learned_parameters, layers,input_shape):
        self.learned_parameters = learned_parameters
        self.layers = layers
        self.outputs = []
        self.input_shape=input_shape
        
        self.total_time = 0
        self.elapsed_time = np.zeros(len(layers))
        gpu_input_memory=None
        
        lp = 0
        last_layer=None
        for layer in self.layers:
            if isinstance(layer, (ConvLayer, DenseLayer)):
                output_shape =layer.load_weights(input_shape = input_shape, weights = self.learned_parameters[lp], biases = self.learned_parameters[lp + 1],enable_gpu=tf.test.is_gpu_available(),gpu_input_memory=gpu_input_memory)
                lp += 2
           
            elif isinstance(layer, (PoolingLayer, FlattenLayer, OutputLayer)):
                output_shape =layer.load_weights(input_shape = input_shape,enable_gpu=tf.test.is_gpu_available(),gpu_input_memory=gpu_input_memory)
                
            else:
                print(f"Layer {type(layer)} not supported yet")
            
            gpu_input_memory=layer.results
            if layer.use_gpu  and   last_layer != None and last_layer.use_gpu:
                last_layer.gpu_output_memory = True
            else:
                print(f"Layer {type(layer)} no gpu")


            input_shape = output_shape
            last_layer = layer
        

    def __deepcopy__(self, memo = None):
        return NetworkModel(self.learned_parameters, copy.deepcopy(self.layers),self.input_shape)

    def export_learned_parameters(self, outdir):
        mkpath(outdir)
        learned_weights = []
        learned_biases  = []
        for i, param in enumerate(grouped(self.learned_parameters, 2)):
            weights = list(flatten(param[0].astype(int).tolist()))
            biases = list(flatten(param[1].astype(int).tolist()))
            with open(f"{outdir}/layer_weights_{i}.json", 'w') as outfile:
                outfile.write(json.dumps(weights))
            with open(f"{outdir}/layer_biases_{i}.json", 'w') as outfile:
                outfile.write(json.dumps(biases))
            learned_weights.append(weights)
            learned_biases.append(biases)
        learned_weights = list(flatten(learned_weights))
        learned_biases = list(flatten(learned_biases))
        with open(f"{outdir}/full_net_weights.json", 'w') as outfile:
            outfile.write(json.dumps(learned_weights)) 
        with open(f"{outdir}/full_net_biases.json", 'w') as outfile:
            outfile.write(json.dumps(learned_biases))

    def predict(self, test_image):
        self.outputs = [test_image]
        for idx, layer in enumerate(self.layers):
            st = time.time()
            if isinstance(layer, (ConvLayer, DenseLayer,PoolingLayer, FlattenLayer, OutputLayer)):
                self.outputs.append(layer.forward_pass(inputv = self.outputs[-1]))
            else:
                print(f"Layer {type(layer)} not supported yet")
            #input_layer.append(outputs)
            et = time.time()
            self.elapsed_time[idx] += et - st
            self.total_time += et - st
        return self.outputs[-1] / np.max(self.outputs[-1])
        
    def print_time_statics(self):
        print(f"Total Time = {self.total_time}")
        for idx, layer in enumerate(self.layers):
            print(f"Layer {type(layer)} # time {self.elapsed_time[idx]}s # ( {self.elapsed_time[idx]/self.total_time*100}% ) #GPU {layer.use_gpu} - GPU_OUT {layer.gpu_output_memory}")
            
    #@staticmethod
    def mimt_do_inference(self, labels, images):
        assert isinstance(labels, (list, tuple))
        assert isinstance(images, (list, tuple))
        assert len(labels) == len(images)
        return [np.array_equal(self.predict(i)[0], l) for l, i in zip(labels, images)]

    #@staticmethod
    def evaluate_accuracy(self, labels, images):
        assert isinstance(labels, (list, tuple))
        assert isinstance(images, (list, tuple))
        assert len(labels) == len(images)
        outputv = self.mimt_do_inference(list(labels),list(images))
        return np.sum(flatten(outputv)) / len(labels) * 100


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
import itertools
from .BaseLayer import *

def mul(a, b, multiplier):
    a = np.array(a)
    b = np.array(b)
    a_shape = np.shape(a)
    b = np.reshape(b, a_shape)
    if multiplier == BaseLayer.default_multiplier:
        return BaseLayer.default_multiplier(a, b)
    return ax_mul(a, b, tuple(a_shape), multiplier)
        
def ax_mul(a, b, shape, multiplier):
    res = np.zeros(shape)
    if len(shape) == 1:
        for i in range(shape[0]):
            res[i] = multiplier.run(int(a[i]),int(b[i]))
    elif len(shape) == 2:
        for i, j in itertools.product(range(shape[0]), range(shape[1])):
            res[i,j] = multiplier.run(int(a[i,j]), int(b[i,j]))
    return res


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
from numba import cuda, int32, int8

@cuda.jit
def max_example_3d(result, values):
    """
    Find the maximum value in values and store in result[0].
    Both result and values are 3d arrays.
    """
    i, j, k = cuda.grid(3)
    # Atomically store to result[0,1,2] from values[i, j, k]
    cuda.atomic.max(result, (0, 1, 2), values[i, j, k])

 
TPB=16 #TODO:calcolare il tpb dinamicamente   
@cuda.jit
def pool(R, inputv, stride, pool_size):#, pooling):
    x, y = cuda.grid(2)
    if x < R.shape[0] and y < R.shape[1]:
        h_start = x * stride[0]
        h_end = h_start + pool_size[0]
        w_start = y * stride[1]
        w_end = w_start + pool_size[1]
        temp_max = cuda.local.array(shape=(TPB), dtype=int32)
        for i in range(h_start,h_end):
            for j in range(w_start,w_end,):
                for k in range(R.shape[2]):
                    if(inputv[i,j,k]>temp_max[k]):
                        temp_max[k]=inputv[i,j,k]
        for k in range(R.shape[2]):
            R[x, y, k]=temp_max[k]

@cuda.jit
def matmul(R, A, B, bias, M, offset1, offset2, activation, quant_nbits):
    """
    Perform matrix di multiplication of C = A * B
    """
    result_max = cuda.shared.array(shape=(8,8), dtype=int32)
    row, col = cuda.grid(2)
    if row < R.shape[0] and col < R.shape[1]:
        tmp = bias[row] #TODO: rendere bias[row,col]
        for k in range(A.shape[1]):
            #tmp +=  M[int8(A[row, k])+offset1 , int(B[k, col])+offset2]
            tmp += A[row, k]*B[k, col]
        #activation rele
        if activation and tmp < 0:
            tmp = 0
        R[row, col] = tmp
    cuda.syncthreads()
    cuda.atomic.max(result_max, (0, 1), R[row, col])
    cuda.syncthreads()
    total_max = result_max[0,0]
    R[row, col] = int8(round(( R[row, col]/total_max)*(2**quant_nbits-1)))
        
@cuda.jit
def convolve(result, a, b, bias, n_channels, M, offset1, offset2, quant_nbits):
    x, y = cuda.grid(2) 
    max_x=a.shape[0]-b.shape[0]+1
    max_y=a.shape[1]-b.shape[1]+1   
    #if the thread coordinates are outside of the image, we ignore the thread:
    if ((x >= max_x)  or (y >= max_y)): 
        return
    #area per fare la max della quantizazione
    #TODO: capire come calcolare lo shape 
    result_max = cuda.shared.array(shape=(4,4,4), dtype=int32)
    #todo: ripassare vettorializazione dati in memoria e assicurarsi che thread dello stesso gruppo usino dati vicini
    for p in range(n_channels):   
        s=bias[p]
        for k in range(a.shape[2]):
            #calcolo cella convoluzione
            #todo: aggiungere lo step
            for i in range(b.shape[0]):
                for j in range(b.shape[1]):
                    s += a[i+x,j+y,k]*b[i,j,k,p]
                    #s += M[int8(a[i+x,j+y,k])+offset1 , int8(b[i,j,k,p])+offset2]
        #TODO supportare altre attivazioni (e.g., softmax, tanh)
        if s < 0:
            s=0
        result[x,y,p] = s
    i, j, k = cuda.grid(3)
    cuda.syncthreads()
    cuda.atomic.max(result_max, (0, 1, 2), result[i, j, k])
    cuda.syncthreads()
    total_max=result_max[0,0,0]
    for p in range(n_channels): 
         result[x,y,p] = int8(round(( result[x,y,p]/total_max)*(2**quant_nbits-1)))
        


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
import numpy as np
from numba import cuda
from .BaseLayer import BaseLayer
from .cuda_utils import convolve

class ConvLayer(BaseLayer):
    def __init__(self, stride = (1, 1), padding = (0,0), activation = "relu", quant_nbits = 8, multiplier = BaseLayer.default_multiplier, name = "layer",offset=[128,128]):
        super().__init__(activation, quant_nbits, multiplier, name)
        self.stride = stride
        self.padding = padding
        self.kernel_global_mem = None
        self.M = None
        self.offset = offset
        
    def __deepcopy__(self, memo = None):
        return ConvLayer(stride = self.stride, padding = self.padding, activation = self.activation, quant_nbits = self.quant_nbits, multiplier = self.multiplier, name = self.name)

    def forward_pass(self, **kwargs):
        if self.gpu_input_memory is None:
            inputv = kwargs["inputv"]
            A_global_mem = cuda.to_device(np.array(inputv))
        else:
            A_global_mem = self.gpu_input_memory 
        convolve[self.griddim, self.blockdim](self.results, A_global_mem,self.kernel_global_mem,self.biases,self.n_channels,self.M,self.offset[0],self.offset[1],self.quant_nbits)
        cuda.synchronize()
        if self.gpu_output_memory == False:
            self.outputv[:,:,:] = self.results.copy_to_host()
            return self.outputv
  
    def load_weights(self, **kwargs):
        self.enable_gpu = kwargs["enable_gpu"]
        self.input_shape= kwargs["input_shape"]
        self.kernel_shape = np.shape(kwargs["weights"])        
        self.output_shape = [int(np.floor((self.input_shape[i] - self.kernel_shape[i] + 2 * self.padding[i]) / self.stride[i] + 1)) for i in range(2)] + [self.kernel_shape[3]]
        self.griddim = (self.input_shape[0] // self.blockdim[0] + 1, self.input_shape[1] // self.blockdim[1] + 1)
        self.n_channels=self.kernel_shape[3]
        if self.enable_gpu:
            self.use_gpu=True
            self.M = self.multiplier
            self.kernel_global_mem = cuda.to_device(np.array(kwargs["weights"]))
            self.biases = cuda.to_device(np.array(kwargs["biases"]))
            self.results = cuda.device_array(self.output_shape)
            self.gpu_input_memory = kwargs["gpu_input_memory"]
        else:
            self.convkernel, self.biases =kwargs["weights"], kwargs["biases"]
            self.outputv = np.zeros(self.output_shape)
        if self.gpu_input_memory is None:
                self.outputv = np.zeros(self.output_shape)
        return self.output_shape
  


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
import numpy as np
from scipy.special import softmax
from functools import partial

class BaseLayer:
    relu = partial(np.maximum, 0)
    max3d = partial(np.max, axis=(0, 1))
    avg3d = partial(np.mean, axis=(0, 1))
    activations = {"relu" : relu, "softmax" : softmax}
    poolings3d = {"max" : max3d, "avg" : avg3d}
    default_multiplier = np.multiply

    def __init__(self, 
            activation = "relu", quant_nbits = 8, multiplier = default_multiplier, name = "layer", 
            enable_gpu = False, gpu_input_memory = None, gpu_output_memory = False, blockdim = (1, 1)):
        self.activation = activation
        self.quant_nbits = quant_nbits
        self.multiplier = multiplier
        self.name = name
        self.weights = None
        self.biases = None
        self.enable_multiprocess = False
        self.enable_gpu = enable_gpu
        self.results = None
        self.use_gpu = False
        self.gpu_output_memory = gpu_output_memory #se è vero non devo passare i dati da results ad outptv
        self.gpu_input_memory = gpu_input_memory  #se è diverso da null non devo spostare i dati in ingresso su gpu (dovrei avere gia l' area di memoria)
        self.blockdim = blockdim # TODO: va calcolato in base alla GPU
        
    def __deepcopy__(self, memo = None):
        return BaseLayer(activation = self.activation, quant_nbits = self.quant_nbits, multiplier = self.multiplier, name = self.name, enablegpu = self.enable_gpu)

    def load_weights(self, **kwargs):
        self.enable_gpu = kwargs["enable_gpu"]
        
        self.weights, self.biases = kwargs["weights"], kwargs["biases"]
            
    def forward_pass(self, **kwargs):
        pass



"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
import numpy as np

def flatten(container):
    for i in container:
        if isinstance(i, (list, tuple)):
            yield from flatten(i)
        else:
            yield i

def grouped(iterable, n):
    return zip(*[iter(iterable)]*n)

def list_partitioning(a_list, num_of_partitions):
    return [list(item) for item in np.array_split(a_list, num_of_partitions)]


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
import numpy as np, itertools
from numba import cuda
from .BaseLayer import BaseLayer
from .cuda_utils import pool

class PoolingLayer(BaseLayer):
    def __init__(self, stride = (1, 1), pool_size = (2, 2), pooling = "max", name = "layer", input_memory = None):
        super().__init__(self, name = name)
        self.stride = stride
        self.pool_size = pool_size
        self.pool = pooling

    def __deepcopy__(self, memo = None):
        return PoolingLayer(stride = self.stride, pool_size = self.pool_size, pooling = self.pool, name = self.name)

    def forward_pass(self, **kwargs):
        if self.gpu_input_memory is None:
            inputv = kwargs["inputv"]
            A_global_mem = cuda.to_device(np.array(inputv))
        else:
            A_global_mem=self.gpu_input_memory
        pool[self.griddim, self.blockdim](self.results, A_global_mem,self.stride_global_mem, self.pool_size_global_mem)#, pooling)
        if self.gpu_output_memory == False:
            cuda.synchronize()
            self.outputv[:,:,:] =self.results.copy_to_host()
            return self.outputv
        cuda.synchronize()

    def load_weights(self, **kwargs):
        self.input_shape, self.enable_gpu = kwargs["input_shape"],kwargs["enable_gpu"] 
        self.output_shape = [int(np.floor((self.input_shape[i] - self.pool_size[i]) / self.stride[i]) + 1) for i in range(2)] + [self.input_shape[2]]      
        self.outputv = np.zeros(self.output_shape)
        self.griddim = (self.output_shape[0] // self.blockdim[0] + 1, self.output_shape[1] // self.blockdim[1] + 1)#,n_channels)
        if self.enable_gpu:
            self.use_gpu=True
            self.results = cuda.device_array(self.output_shape)
            self.stride_global_mem = cuda.to_device(self.stride)
            self.pool_size_global_mem = cuda.to_device(self.pool_size)
            self.gpu_input_memory = kwargs["gpu_input_memory"]
        return self.output_shape
    
    def pooling_kernel(self, inputv, pooling):
        function_pool=BaseLayer.poolings3d[pooling]
        inner_range = self.output_shape[1]
        outer_iterations=range(self.output_shape[1])
        for i, j in itertools.product(outer_iterations, range(inner_range)):
            h_start = i * self.stride[0]
            h_end = h_start + self.pool_size[0]
            w_start = j * self.stride[1]
            w_end = w_start + self.pool_size[1]
            self.outputv[i, j, :] = function_pool(inputv[h_start:h_end, w_start:w_end, :])


from .ConvLayer import ConvLayer
from .DenseLayer import DenseLayer
from .PoolingLayer import PoolingLayer
from .FlattenLayer import FlattenLayer
from .OutputLayer import OutputLayer
from .NetworkModel import NetworkModel
name = "inspectnn"
__version__ = "0.0.1"
__author__ = "Salvatore Barone, Filippo Ferrandino"
__credits__ = "Department of Electrical Engineering and Information Technologies, University of Naples Federico II, Via Claudio 21, Naples, Italy"



"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
import numpy as np
from numba import cuda
from .BaseLayer import BaseLayer

class FlattenLayer(BaseLayer):
    def __init__(self, name = "layer"):
        super().__init__(self, name = name)

    def __deepcopy__(self, memo = None):
        return FlattenLayer(name = self.name)

    def load_weights(self, **kwargs):
        self.input_shape, self.enable_gpu = kwargs["input_shape"], kwargs["enable_gpu"] 
        self.output_shape = [1, np.prod(self.input_shape)]
        self.outputv = np.zeros(self.output_shape)
        if self.enable_gpu:
            self.use_gpu = True
            self.gpu_input_memory = kwargs["gpu_input_memory"]
        return self.output_shape
    
    def forward_pass(self, **kwargs):
        if self.gpu_input_memory is None:
            inputv = kwargs["inputv"]
            print("flat no buoeno")
            return np.reshape(inputv, self.output_shape)
        else:
            # TODO reshape su GPU
            self.results = self.gpu_input_memory.reshape(self.output_shape[0],self.output_shape[1])
            cuda.synchronize()
            self.outputv[:,:]=self.results.copy_to_host()
            return self.outputv

        